{$M+}
PROGRAM bnfout;

 CONST  MAX_PROD  = 511;
        PROLEN    = 30;
        TP        = FALSE;
        GP        = TRUE;
        TEXT_S    = '.TEXT';
        DATA_S    = '.DATA';
        GO_TAB    = 'GO_TAB:';
        GOTO_     = 'GOTO:';

 TYPE   str255     = string[255];
        s255ptr    = ^str255;
{ s255ptr hat dynamisch nicht unbedingt 255 Zeichen -> Lnge beachten }

{ Mit folgenden Deklarationen wird ein Zugriff auf Strings ermglicht, die
  die Routinen GET_ACTION, ... enthalten. write_data schreibt die passenden
  Routinen in die Ausgabedatei. Die Strings stehen im anschlieenden
  Assemblertext (GFA) }

        rout_arr   = ARRAY [boolean] OF s255ptr;
        rout_rec   = RECORD
                      pro : s255ptr;
                      lshlp : s255ptr;
                      acts: ARRAY [boolean] OF rout_arr;
                      gos : ARRAY [boolean,0..1] OF rout_arr
                     END;
        rout_ptr   = ^rout_rec;

{ In den Strukturen des_tab und ter_tab werden Informationen zu Nonterminalen
  bzw. Terminalen gespeichert. dlen und tlen geben Auskunft ber deren Anzahl.
  Bei Terminalen wird nur der Name festgehalten. In weiteren Berechnungen wer-
  den Terminale durch ihren negativen, Nonterminale dagegen durch positiven
  Index reprsentiert. d_entry beinhaltet neben dem Namen auch einen Zeiger
  auf eine Liste, deren Glieder r_side die rechte Seite von Produktionen, die
  das Nonterminal lexeme^ als linke Seite haben, darstellen. Die Zahlen im
  Array buf sind Indizes von des_tab.tab und ter_tab.lex. }

        r_side_ptr = ^r_side;
        r_side     = RECORD
                      pno :integer;     { Nummer der Produktion }
                      npno:integer;     { Nummer der Produktion in BNF }
                      des :integer;     { Index des Nont. auf rechter Seite }
                      blen:byte;        { Lnge von buf (der r. Seite ) }
                      next:r_side_ptr;  { Nchste Produktion }
                      cons:s255ptr;     { Name, in der BNF def. }
                      buf :ARRAY [1..PROLEN] OF integer
                     END;
        prod_tab   = ARRAY [0..MAX_PROD] OF integer;
        d_entry    = RECORD
                      lexeme :s255ptr;      { Name Nonterminal }
                      prolist:r_side_ptr;   { r. Seiten }
                      res1   :long_integer; { wird von Ausgaberoutinen nicht }
                      res2   :long_integer; { gebraucht. }
                     END;
        des_tab    = RECORD
                      dlen:integer;
                      tab :ARRAY [0..MAX_PROD] OF d_entry
                     END;
        ter_tab    = RECORD
                      tlen:integer;
                      lex :ARRAY [-MAX_PROD..-1] OF s255ptr
                     END;
        comp_ptr   = ^valid_comp;
        valid_comp = RECORD
                      index : integer;
                      state : integer;
                      next  : comp_ptr;
                     END;
        n_list     = ARRAY [0..MAX_PROD] OF comp_ptr;

{ Zu jedem state des DFA gehrt eine Variable des Typs kernel. Jeder state hat
  Eintrge in ACT_TAB und GO_TAB. ACT_TAB wird mittels state und token, GO_TAB
  mittels state und Nonterminal adressiert (ACT_TAB[state,token] und
  GO_TAB[state,Nonterminal]). Da in den Tabellen des fteren gleiche Zeilen
  vorkommen, werden alle unterschiedlichen in Listen gespeichert. Die Zeile
  eines states kann ber tabs[TRUE/FALSE]^ erreicht werden. GOTO[state] ist
  dabei tabs[TRUE]^. Viele Eintrge einer Zeile sind null; deswegen sind die
  Zeilen selbst auch Listen, ein Eintrag bestehend aus Index und Wert. }

        st_ptr     = ^st_rec;
        st_rec     = RECORD
                      next:st_ptr;
                      resv:long_integer;
                      stno:long_integer;
                      stb :comp_ptr
                     END;
        ker_ptr    = ^kernel;
        kernel     = RECORD
                      next:ker_ptr;
                      tabs:ARRAY [boolean] OF st_ptr
                     END;

 FUNCTION get_rout:rout_ptr; EXTERNAL;

{ write_pr gibt ein Item in lesbarer Form aus. d ist die Position des
  Punktes. Fr d < 0 wird eine Produktion ausgegeben. Vom Hauptprogramm wird
  write_pr benutzt, mu deshalb global sein. }

{$E+}
 PROCEDURE write_pr(VAR des_rec:des_tab; VAR term_rec:ter_tab;
                    VAR p:r_side; d:integer);

  VAR i : integer;

  BEGIN
   WITH p, des_rec, term_rec DO
    BEGIN
     write(-pno,#9,tab[des].lexeme^,#9,': ');
     FOR i:= 1 TO blen DO
      BEGIN
       IF i = d THEN write('.');
       IF buf[i] < 0 THEN write(lex[buf[i]]^,' ')
                     ELSE write(tab[buf[i]].lexeme^,' ')
      END;
     IF d > blen THEN write('.');
     writeln
    END
  END;
{$E-}

{ write_prod schreibt die Tabelle pr_tab }

 PROCEDURE write_prod(VAR des_rec:des_tab; VAR term_rec:ter_tab);

  VAR i,no : integer;
      p : r_side_ptr;
      s : string[6];

  BEGIN
   s:= '.DC.b ';
   s[6]:= #9;
   IF des_rec.dlen > 255 THEN s[4]:= 'w'; { Mehr als 255 Nonterminale }

   { Fr alle Produktionen ... }

   FOR i:= 1 TO des_rec.dlen DO
    BEGIN
     p:= des_rec.tab[i].prolist;
     WHILE p <> NIL DO
      BEGIN
       write(s,pred(i),',',p^.blen,#9,';');
       write_pr(des_rec,term_rec,p^,-1);
       p:= p^.next
      END
    END
  END;

{ wr_tab gibt die Zeile eines states aus. Entweder ACT_TAB[no] oder
  GO_TAB[no]. }

 PROCEDURE wr_tab(p:comp_ptr; s,e,no:integer; c:char; list:boolean);

  VAR i,open,state,lh : integer;
      str             : ARRAY [boolean] OF string;

  BEGIN
   write(';',no);
   open:= 0;
   IF list THEN c:= 'w';
   str[TRUE]:= '  .DC.  ';
   str[TRUE][1]:= #13; str[TRUE][2]:= #10;
   str[TRUE][7]:= c; str[TRUE][8]:= #9;
   str[FALSE]:= ',';
   IF list THEN
    BEGIN
     WHILE p <> NIL DO
      BEGIN
       write(str[open = 0],p^.index,',',p^.state);
       open:= (open + 2) & 7;
       p:= p^.next
      END;
     write(str[open = 0],'-1')
    END    ELSE
    BEGIN
     IF p = NIL THEN lh:= succ(e) ELSE lh:= p^.index;
     FOR i:= s TO e DO
      BEGIN
       IF i <> lh THEN state:= 0 ELSE
        BEGIN
         state:= p^.state;
         p:= p^.next;
         IF p = NIL THEN lh:= succ(e) ELSE lh:= p^.index
        END;
       write(str[open = 0],state);
       open:= succ(open) & 7
      END
    END;
   writeln
  END;

{ buf_out schreibt alle unterschiedlichen Zeile von ACT_TAB oder GO_TAB.
  write_data kennt drei Ausgabeformate und verwendet dasjenige, das die
  Tabelle am komprimiertesten ausgibt:
  1. Tabelle ohne Offset
  2. Tabelle mit Offset
  3. Tabelle als Liste -> keine Nulleintrge }

 PROCEDURE buf_out(sp:st_ptr; s,e:integer; msg:string; c:char; list:boolean);

  BEGIN
   writeln(msg);
   WHILE sp <> NIL DO
    WITH sp^ DO
     BEGIN
      writeln(msg[1],stno:4:h,':');
      wr_tab(stb,s,e,stno,c,list);
      sp:= sp^.next
    END
  END;

{ ACT_TAB und GO_TAB ausgeben }

 PROCEDURE ptable_out(p:ker_ptr; str:string; width:char; com,ga:boolean;
                      s,e:integer; sno:long_integer);

  VAR i : long_integer;

  BEGIN
   writeln(str);
   FOR i:= 0 TO sno DO
    BEGIN
     IF com THEN writeln('.DC.l  ',str[1],p^.tabs[ga]^.stno:4:h,';':4,i)
            ELSE wr_tab(p^.tabs[ga]^.stb,s,e,i,width,FALSE);
       p:= p^.next
    END
  END;

 FUNCTION size(len,sno,dno,entries:long_integer; asl:integer;
               op:char; VAR offs,list:boolean):boolean;

  VAR t1,t2,t3 : long_integer;

  BEGIN
   t1:= shl(sno * len,asl);                 { Tabelle ohne Offset }
   t2:= shl(sno,2) + shl(dno * len,asl);    { Tabelle mit Offset }
   t3:= shl(shl(sno + entries,1) + dno,1);  { Liste mit Offset }
   list:= (t2 > t3) AND (op <> 'q');
   offs:= (t2 < t1) OR ((t3 < t1) AND list);
   size:= list AND offs
  END;

{ write_data ruft das Hauptprogramm auf. Parameter:
  p   : Liste der States
  sno : Lnge von p
  gno : Lnge von go_
  ano : Lnge von act_
  act_: Liste aller unterschiedlichen ACT_TAB-zeilen
  go_ : Liste aller unterschiedlichen GO_TAB-zeilen
  act_entries : Anzahl Eintrge in act_ <> 0
  go_entries : Anzahl Eintrge in go_ <> 0
  op : optionaler Parameter aus Kommandozeile }

{$E+}
 PROCEDURE write_data (p:ker_ptr; sno,gno,ano:long_integer; act_,go_:st_ptr;
                       act_entries,go_entries:long_integer; op:char;
                       VAR des_rec:des_tab; VAR term_rec:ter_tab);

  VAR offs,list,lsout : boolean;
      ch : char;
      rout : rout_ptr;
      word : integer;

  BEGIN
   word:= ord(sno > 255);
   IF word = 1 THEN ch:= 'w' ELSE ch:= 'b';
   rout:= get_rout;

   WITH des_rec, term_rec, rout^ DO
    BEGIN
     lsout:= size(succ(-tlen),succ(sno),succ(ano),act_entries,1,op,offs,list);
     IF NOT offs THEN writeln('ACT_LEN equ ',shl(-pred(tlen),1));
     writeln(TEXT_S,#13,#10,pro^,acts[offs][list]^,
             DATA_S,#13,#10,'pr_tab:');
     write_prod(des_rec,term_rec);
     ptable_out(p,'ACT_TAB:','w',offs,TP,0,-tlen,sno);
     IF offs THEN buf_out(act_,0,-tlen,'ACTION:','w',list);

     { offs = TRUE, wenn ber Offsettabelle auf einen state zugegriffen
       wird }

     lsout:= size(dlen,succ(sno),succ(gno),go_entries,word,op,offs,list)
             OR lsout;
     IF NOT offs THEN writeln('GO_LEN equ ',shl(dlen,word));
     write(TEXT_S,#13,#10,gos[offs][word][list]^);
     IF lsout THEN write(lshlp^);
     writeln(DATA_S);
     ptable_out(p,GO_TAB,ch,offs,GP,1,dlen,sno);
     IF offs THEN buf_out(go_,1,dlen,GOTO_,ch,list)
    END
  END;
{$E-}
BEGIN END.
{ Assemblierter Rest der Datei in bnfoasm.o }
GET_ROUT::
          move.l    #routines,d0
          rts
          .DATA
routines:
          .DC.l pro_r
          .DC.l list
          .DC.l act_no,act_no,act_com,act_lst
          .DC.l go_nob,go_nob
          .DC.l go_now,go_now
          .DC.l go_comb,go_lst
          .DC.l go_comw,go_lst
pro_r:
          .DC.b (act_com-*)-1
          .DC.b "GET_PROD::",13,10
          .DC.b 9,"move.l",9,"#pr_tab,d0",13,10
          .DC.b 9,"rts",13,10
act_com:
          .DC.b (act_no-*)-1
          .DC.b "GET_ACTION::",13,10
          .DC.b 9,"movea.l",9,"(sp)+,a0  ;Return",13,10
          .DC.b 9,"move.w",9,"(sp)+,d0  ;token",13,10
          .DC.b 9,"add.w",9,"d0,d0     ;integer",13,10
          .DC.b 9,"move.l",9,"(sp)+,d1  ;state",13,10
          .DC.b 9,"lsl.l",9,"#2,d1     ;*4",13,10
          .DC.b 9,"lea",9,"ACT_TAB,a1",13,10
          .DC.b 9,"movea.l",9,"0(a1,d1.l),a1",13,10
          .DC.b 9,"move.w",9,"0(a1,d0.w),d0",13,10
          .DC.b 9,"jmp",9,"(a0)",13,10
act_no:
          .DC.b (act_lst-*)-1
          .DC.b "GET_ACTION::",13,10
          .DC.b 9,"movea.l",9,"(sp)+,a0  ;Return",13,10
          .DC.b 9,"move.w",9,"(sp)+,d0  ;token",13,10
          .DC.b 9,"add.w",9,"d0,d0     ;integer",13,10
          .DC.b 9,"move.l",9,"(a7)+,d1  ;<2*maxint",13,10
          .DC.b 9,"mulu.w",9,"#ACT_LEN,d1",13,10
          .DC.b 9,"lea",9,"ACT_TAB,a1",13,10
          .DC.b 9,"adda.l",9,"d1,a1",13,10
          .DC.b 9,"move.w",9,"0(a1,d0.w),d0",13,10
          .DC.b 9,"jmp",9,"(a0)",13,10
act_lst:
          .DC.b (go_comb-*)-1
          .DC.b "GET_ACTION::",13,10
          .DC.b 9,"lea",9,"ACT_TAB,a1",13,10
          .DC.b 9,"bra",9,"list",13,10
go_comb:
          .DC.b (go_nob-*)-1
          .DC.b "GET_GOTO::",13,10
          .DC.b 9,"movea.l",9,"(sp)+,a0  ;Return",13,10
          .DC.b 9,"move.w",9,"(sp)+,d2  ;Nonterm",13,10
          .DC.b 9,"move.l",9,"(sp)+,d1  ;state",13,10
          .DC.b 9,"lsl.l",9,"#2,d1     ;*4",13,10
          .DC.b 9,"lea",9,"GO_TAB,a1",13,10
          .DC.b 9,"movea.l",9,"0(a1,d1.l),a1",13,10
          .DC.b 9,"clr.w",9,"d0",13,10
          .DC.b 9,"move.b",9,"0(a1,d2.w),d0",13,10
          .DC.b 9,"jmp",9,"(a0)",13,10
go_nob:
          .DC.b (go_comw-*)-1
          .DC.b "GET_GOTO::",13,10
          .DC.b 9,"movea.l",9,"(sp)+,a0  ;Return",13,10
          .DC.b 9,"move.w",9,"(sp)+,d2  ;Nonterm",13,10
          .DC.b 9,"move.l",9,"(a7)+,d1  ;<2*maxint",13,10
          .DC.b 9,"mulu.w",9,"#GO_LEN,d1",13,10
          .DC.b 9,"lea",9,"GO_TAB,a1",13,10
          .DC.b 9,"adda.l",9,"d1,a1",13,10
          .DC.b 9,"clr.w",9,"d0",13,10
          .DC.b 9,"move.b",9,"0(a1,d2.w),d0",13,10
          .DC.b 9,"jmp",9,"(a0)",13,10
go_comw:
          .DC.b (go_now-*)-1
          .DC.b "GET_GOTO::",13,10
          .DC.b 9,"movea.l",9,"(sp)+,a0  ;Return",13,10
          .DC.b 9,"move.w",9,"(sp)+,d0  ;nonterm",13,10
          .DC.b 9,"add.w",9,"d0,d0     ;integer",13,10
          .DC.b 9,"move.l",9,"(sp)+,d1  ;state",13,10
          .DC.b 9,"lsl.l",9,"#2,d1     ;*4",13,10
          .DC.b 9,"lea",9,"GO_TAB,a1",13,10
          .DC.b 9,"movea.l",9,"0(a1,d1.l),a1",13,10
          .DC.b 9,"move.w",9,"0(a1,d0.w),d0",13,10
          .DC.b 9,"jmp",9,"(a0)",13,10
go_now:
          .DC.b (go_lst-*)-1
          .DC.b "GET_GOTO::",13,10
          .DC.b 9,"movea.l",9,"(sp)+,a0  ;Return",13,10
          .DC.b 9,"move.w",9,"(sp)+,d0  ;token",13,10
          .DC.b 9,"add.w",9,"d0,d0     ;integer",13,10
          .DC.b 9,"move.l",9,"(sp)+,d1  ;<2*maxint",13,10
          .DC.b 9,"mulu.w",9,"#GO_LEN,d1",13,10
          .DC.b 9,"lea",9,"GO_TAB,a1",13,10
          .DC.b 9,"adda.l",9,"d1,a1",13,10
          .DC.b 9,"move.w",9,"0(a1,d0.w),d0",13,10
          .DC.b 9,"jmp",9,"(a0)",13,10
go_lst:
          .DC.b (list-*)-1
          .DC.b "GET_GOTO::",13,10
          .DC.b 9,"lea",9,"GO_TAB,a1",13,10
list:
          .DC.b (finito-*)-1
          .DC.b "list:",13,10
          .DC.b 9,"movea.l",9,"(sp)+,a0",13,10
          .DC.b 9,"move.w",9,"(sp)+,d2",13,10
          .DC.b 9,"move.l",9,"(sp)+,d1",13,10
          .DC.b 9,"lsl.l",9,"#2,d1",13,10
          .DC.b 9,"movea.l",9,"0(a1,d1.l),a1",13,10
          .DC.b 9,"clr.w",9,"d0",13,10
          .DC.b ".1:",13,10
          .DC.b 9,"cmp.w",9,"(a1)+,d2",13,10
          .DC.b 9,"bhi",9,".2",13,10
          .DC.b 9,"bne",9,".3",13,10
          .DC.b 9,"move.w",9,"(a1),d0",13,10
          .DC.b ".3:",9,"jmp",9,"(a0)",13,10
          .DC.b ".2:",13,10
          .DC.b 9,"addq.l",9,"#2,a1",13,10
          .DC.b 9,"bra",9,".1",13,10
finito:

