Program Slr;

{  Syntaxfree grammar Parser
   1  G -> R
   2  R -> R o C
   3  R -> C
   4  C -> C S
   5  C -> S
   6  S -> F *
   7  S -> F
   8  F -> ( R )
   9  F -> l
   10 F -> [ I ]
   11 I -> I l
   12 I -> I l - l
   13 I -> l
   14 I -> l - l      }

 const maxpos    = 255;
       maxnodes  = 255;
       reg       = 0;
       conc      = 1;
       clos      = 2;
       fak       = 3;
       range     = 4;
       R_RoC  = 4;
       C_CS   = 6;
       S_Fcl  = 8;
       F_I    = 9;
       F_l    = 10;
       F_R    = 11;
       I_l_l  = 12;
       I_l    = 13;
       I_Il_l = 14;
       I_Il   = 15;

 type token  = (eos,o,c,bo,bc,l,bro,brc,ra);
               {eof,|,*,( ,) ,l,[  ,]  ,- }
      pr_typ   = ARRAY [2..15] OF
                  PACKED RECORD
                   pr :byte;
                   len:byte
                  END;
      prod_ptr = ^pr_typ;
      stack  = record
                state:packed array [byte] of byte;
                val  :packed array [byte] of byte;
                lex  :packed array [byte] of char
               end;
      posset = packed set of 1..maxpos;
      setstk = array [1..maxnodes] of posset;
      null   = packed array [0..maxnodes] of boolean;
      str255    = string[255];

 var st        :stack;
     s         :str255;
     leafno,setcount,i,j,ip:integer;
     fpos,lpos,fwpos:setstk;
     nullable  :null;
     lexval    :char;
     symtab    :str255;
     symbols   :str255;
     include   :set of char;

 FUNCTION get_prod:prod_ptr; EXTERNAL;
 FUNCTION get_action(state:long_integer;t:token):integer; EXTERNAL;
 FUNCTION get_goto(state:long_integer;n:integer):integer; EXTERNAL;

 Procedure appchar(var s:str255;c:char);

  BEGIN
   s[0]:=succ(s[0]);
   s[length(s)]:=c
  END;

 Function upcase(c:char):char;

  BEGIN
   IF c IN ['a'..'z'] THEN c:=chr(ord(c) - 32);
   upcase:=c
  END;

 Procedure augment(var s:str255);

  var i:integer;

  BEGIN
   s[0]:=succ(s[0]);
   FOR i:=length(s) DOWNTO 2 DO s[i]:=s[pred(i)];
   s[1]:='(';
   appchar(s,')');
   appchar(s,#255)
  END;

 Procedure get_symbols(var so:str255);

  var c:char;

  BEGIN
   so:='';
   FOR c:=#1 TO #254 DO IF c IN include THEN appchar(so,c)
  END;

 Function get_token(var str:str255):token;

  BEGIN
   case str[ip] of
    '*','+','?':
        BEGIN
         get_token:=c;
         lexval:=str[ip]
        END;
    '\':BEGIN
         ip:=succ(ip);
         lexval:=str[ip];
         get_token:=l
        END;
    '-':get_token:=ra;
    '|':get_token:=o;
    '[':get_token:=bro;
    ']':get_token:=brc;
    '(':get_token:=bo;
    ')':get_token:=bc;
    #0 :get_token:=eos;
    otherwise:BEGIN
               get_token:=l;
               lexval:=str[ip]
              END
   end
  END;

 Procedure makeleaf(no:integer;c:char);

  BEGIN
   leafno:=succ(leafno);
   nullable[no]:=false;
   fpos[no]:=[leafno];
   lpos[no]:=[leafno];
   symtab[leafno]:=c;
   include:=include + [c];
   fwpos[leafno]:=[]
  END;

 Procedure emptyleaf(n:integer);

  BEGIN
   nullable[n]:=true;
   fpos[n]:=[];
   lpos[n]:=[]
  END;

 Procedure ornode(n,c1,c2:integer);

  BEGIN
   nullable[n]:=nullable[c1] or nullable[c2];
   fpos[n]:=fpos[c1] + fpos[c2];
   lpos[n]:=lpos[c1] + lpos[c2]
  END;

 Procedure concatnode(n,c1,c2:integer);

  var i:integer;

  BEGIN
   FOR i:=1 TO leafno DO IF i IN lpos[c1] THEN fwpos[i]:=fwpos[i] + fpos[c2];
   IF nullable[c1] THEN fpos[n]:=fpos[c1] + fpos[c2]
                   ELSE fpos[n]:=fpos[c1];
   IF nullable[c2] THEN lpos[n]:=lpos[c1] + lpos[c2]
                   ELSE lpos[n]:=lpos[c2];
   nullable[n]:=nullable[c1] and nullable[c2]
  END;

 Procedure closurenode(n,c:integer);

  var i:integer;

  BEGIN
   nullable[n]:=true;
   fpos[n]:=fpos[c];
   lpos[n]:=lpos[c];
   FOR i:=1 TO leafno DO IF i IN lpos[n] THEN fwpos[i]:=fwpos[i] + fpos[n]
  END;

 Function calc_pos(pr,sp:integer):boolean;

  var c:char;

  BEGIN
   WITH st DO
    case pr of
     F_l,I_l:BEGIN
              setcount:=succ(setcount);
              makeleaf(setcount,lex[sp])
             END;
     R_RoC  :BEGIN
              setcount:=val[sp - 2];
              ornode(setcount,setcount,val[sp])
             END;
     I_Il   :BEGIN
              makeleaf(succ(setcount),lex[sp]);
              ornode(setcount,setcount,succ(setcount))
             END;
     I_Il_l,
     I_l_l   :BEGIN
              IF pr = I_l_l THEN
               BEGIN
                setcount:=succ(setcount);
                lpos[setcount]:=[];
                fpos[setcount]:=[];
                nullable[setcount]:=false
               END;
              FOR c:=lex[sp - 2] TO lex[sp] DO
               BEGIN
                makeleaf(succ(setcount),c);
                ornode(setcount,setcount,succ(setcount))
               END
             END;
     C_CS   :BEGIN
              setcount:=val[sp - 1];
              concatnode(setcount,val[sp - 1],val[sp])
             END;
     S_Fcl  :BEGIN
              IF (lex[sp] = '*') or (lex[sp] = '+') THEN
               BEGIN
                closurenode(succ(setcount),val[sp - 1]);
                IF lex[sp] = '+' THEN
                  concatnode(succ(setcount),val[sp - 1],succ(setcount));
                fpos[setcount]:=fpos[succ(setcount)];
                lpos[setcount]:=lpos[succ(setcount)];
                nullable[setcount]:=nullable[succ(setcount)]
               END                                  ELSE
               BEGIN
                emptyleaf(succ(setcount));
                ornode(setcount,val[sp - 1],succ(setcount))
               END
             END
    end;
   calc_pos:=(leafno <= maxpos)
  END;

 Function parse(str:str255):boolean;

  var t             :token;
      sp,sp2,s,pr   :integer;
      not_acc,no_err:boolean;
      production    :prod_ptr;

  BEGIN
   str[succ(length(str))]:=#0;
   ip:=1;
   sp:=0;
   leafno:=0;
   production:= get_prod;
   WITH st DO
    BEGIN
     state[sp]:=0;
     t:=get_token(str);
     not_acc:=true;
     no_err:=true;
     WHILE no_err and not_acc DO
      BEGIN
       s:=state[sp];
       pr:= get_action(s,t);
       IF pr > 0 THEN
        BEGIN
         sp:=succ(sp);
         ip:=succ(ip);
         state[sp]:= pr;
         lex[sp]:=lexval;
         t:=get_token(str)
        END               ELSE
       IF pr < -1 THEN
        BEGIN
         pr:= -pr;
         sp2:=sp - pred(production^[pr].len);
         case pr of
          R_RoC,C_CS,S_Fcl,F_l,I_Il,I_l,I_Il_l,I_l_l:
           BEGIN
            no_err:=calc_pos(pr,sp);
            val[sp2]:=setcount
           END;
          F_R,F_I:val[sp2]:=val[pred(sp)]
         end;
         sp:=sp2;
         state[sp]:= get_goto(state[pred(sp)],production^[pr].pr)
        END               ELSE
       BEGIN
        not_acc:= (pr <> -1);
        no_err := NOT not_acc
       END
      END
    END;
   IF not_acc THEN writeln('Fehler !!!',ord(t):4,s:5)
              ELSE writeln('Erfolg !!!');
   parse:=no_err
  END;

 Procedure cr_trtab;

  var i,k,m,checked:integer;
      sets         :setstk;
      cset         :posset;
      bool         :boolean;

  BEGIN
   m:=1;
   sets[1]:=fpos[1];
   checked:=1;
   write(' ':5);
   FOR i:=1 TO length(symbols) DO write(symbols[i]:3,'  ');
   writeln;
   WHILE checked <= m DO
    BEGIN
     write(pred(checked):4,' ');
     FOR i:=1 TO length(symbols) DO
      BEGIN
       cset:=[];
       FOR k:=1 TO leafno DO
        IF k IN sets[checked] THEN
         IF symtab[k] = symbols[i] THEN cset:=cset + fwpos[k];
       IF cset = [] THEN write(' ':5)
                    ELSE
        BEGIN
         k:=1;
         bool:=true;
         WHILE bool and (k <= m) DO
          BEGIN
           bool:=(sets[k] <> cset);
           k:=succ(k)
          END;
         IF bool THEN
          BEGIN
           sets[k]:=cset;
           m:=succ(m)
          END;
         write(pred(k) - ord(not bool):3,'  ')
        END
      END;
     checked:=succ(checked);
     writeln
    END;
   writeln('Acc. States:');
   FOR i:=1 TO m DO IF leafno IN sets[i] THEN write(pred(i):3);
   writeln
  END;

 BEGIN
  LOOP
   include:=[];
   setcount:=0;
   write('->');readln(s);
   Exit IF s = '';
   augment(s);
   IF parse(s) THEN
    BEGIN
     get_symbols(symbols);
     cr_trtab
    END
  END
 END.
ACT_LEN equ 18
GET_PROD::
          move.l    #pr_tab,d0
          rts
GET_ACTION::
        movea.l   (sp)+,a0  ;Return
        move.w    (sp)+,d0  ;token
        add.w     d0,d0     ;integer
        move.l    (a7)+,d1  ;<2*maxint
        mulu.w    #ACT_LEN,d1
        lea       ACT_TAB,a1
        adda.l    d1,a1
        move.w    0(a1,d0.w),d0
        jmp       (a0)
GET_GOTO::
        movea.l   (sp)+,a0  ;Return
        move.w    (sp)+,d0  ;Nonterm
        move.l    (sp)+,d1  ;state
        lsl.l     #2,d1     ;*4
        lea       GO_TAB,a1
        movea.l   0(a1,d1.l),a1
        move.b    0(a1,d0.w),d0
        ext.w     d0
        jmp       (a0)
pr_tab:
DC.b    0,1     ;2      G       : R
DC.b    1,1     ;3      R       : C
DC.b    1,3     ;4      R       : R o C
DC.b    2,1     ;5      C       : S
DC.b    2,2     ;6      C       : C S
DC.b    3,1     ;7      S       : F
DC.b    3,2     ;8      S       : F *
DC.b    4,3     ;9      F       : [ I ]
DC.b    4,1     ;10     F       : l
DC.b    4,3     ;11     F       : ( R )
DC.b    5,3     ;12     I       : l - l
DC.b    5,1     ;13     I       : l
DC.b    5,4     ;14     I       : I l - l
DC.b    5,2     ;15     I       : I l
ACT_TAB:
;0
DC.w         0,0,0,6,0,7,8,0,0
;1
DC.w        -7,-7,9,-7,-7,-7,-7,0,0
;2
DC.w        -5,-5,0,-5,-5,-5,-5,0,0
;3
DC.w        -3,-3,0,6,-3,7,8,0,0
;4
DC.w        -2,11,0,0,0,0,0,0,0
;5
DC.w        -1,0,0,0,0,0,0,0,0
;6
DC.w         0,0,0,6,0,7,8,0,0
;7
DC.w       -10,-10,-10,-10,-10,-10,-10,0,0
;8
DC.w         0,0,0,0,0,14,0,0,0
;9
DC.w        -8,-8,0,-8,-8,-8,-8,0,0
;10
DC.w        -6,-6,0,-6,-6,-6,-6,0,0
;11
DC.w         0,0,0,6,0,7,8,0,0
;12
DC.w         0,11,0,0,16,0,0,0,0
;13
DC.w         0,0,0,0,0,17,0,18,0
;14
DC.w         0,0,0,0,0,-13,0,-13,19
;15
DC.w        -4,-4,0,6,-4,7,8,0,0
;16
DC.w       -11,-11,-11,-11,-11,-11,-11,0,0
;17
DC.w         0,0,0,0,0,-15,0,-15,20
;18
DC.w        -9,-9,-9,-9,-9,-9,-9,0,0
;19
DC.w         0,0,0,0,0,21,0,0,0
;20
DC.w         0,0,0,0,0,22,0,0,0
;21
DC.w         0,0,0,0,0,-12,0,-12,0
;22
DC.w         0,0,0,0,0,-14,0,-14,0
GO_TAB:
DC.l  G000000   ;0
DC.l  G000001   ;1
DC.l  G000001   ;2
DC.l  G000002   ;3
DC.l  G000001   ;4
DC.l  G000001   ;5
DC.l  G000003   ;6
DC.l  G000001   ;7
DC.l  G000004   ;8
DC.l  G000001   ;9
DC.l  G000001   ;10
DC.l  G000005   ;11
DC.l  G000001   ;12
DC.l  G000001   ;13
DC.l  G000001   ;14
DC.l  G000002   ;15
DC.l  G000001   ;16
DC.l  G000001   ;17
DC.l  G000001   ;18
DC.l  G000001   ;19
DC.l  G000001   ;20
DC.l  G000001   ;21
DC.l  G000001   ;22
GOTO:
G000000:
;0
DC.b         5,4,3,2,1,0
G000001:
;1
DC.b         0,0,0,0,0,0
G000002:
;2
DC.b         0,0,0,10,1,0
G000003:
;3
DC.b         0,12,3,2,1,0
G000004:
;4
DC.b         0,0,0,0,0,13
G000005:
;5
DC.b         0,0,15,2,1,0

